#include "common_header.h"
#include "OpenGLControl.h"

/*-----------------------------------------------

Name:   InitGLEW

Params: none

Result: Creates fake window and OpenGL rendering
        context, within which GLEW is initialized.

/*---------------------------------------------*/

bool COpenGLControl::InitGLEW()
{
	if (bGlewInitialized)
		return true;

	bool bResult = true;
	if(!bGlewInitialized)
	{
		glewExperimental = true;
		if(glewInit() != GLEW_OK)
		{
			PRINT_ERROR("Couldn't initialize GLEW!");
			bResult = false;
		}
		glGetError(); // Discard GLEW errors
		bGlewInitialized = true;
	}

	return bResult;
}

/*-----------------------------------------------

Name:   InitOpenGL

Params: a_window - window to init OpenGL into
        a_initScene - pointer to init function
        a_renderScene - pointer to render function
        a_releaseScene - optional parameter of release function

Result: Initializes OpenGL rendering context
        of specified version. If succeeds, returns true.

/*---------------------------------------------*/

bool COpenGLControl::InitOpenGL(SDL_Window * a_window, void (*a_InitScene)(void *), void (*a_RenderScene)(void *), void(*a_ReleaseScene)(void *), void * lpParam)
{
	window = a_window;
	context = SDL_GL_CreateContext(window);
	if (!context) {
		PRINT_ERROR("Error creating OpenGL Context");
		return false;
	}
	SDL_GL_SetSwapInterval(1);

	PRINT_INFO("OpenGL Version: " << glGetString(GL_VERSION));

	if (!InitGLEW())
		return false;

	RenderScene = a_RenderScene;
	InitScene = a_InitScene;
	ReleaseScene = a_ReleaseScene;

	if (InitScene != NULL)
		InitScene(lpParam);

	return true;
}

/*-----------------------------------------------

Name:   InitSDL2

Params: a_iMajorVersion - Major version of OpenGL
        a_iMinorVersion - Minor version of OpenGL

Result: Sets various hints to let SDL know what
        options we want in our OpenGL window

/*---------------------------------------------*/

bool COpenGLControl::InitSDL2(int iMajorVersion, int iMinorVersion)
{
	SDL_Init(SDL_INIT_VIDEO);

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, iMajorVersion);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, iMinorVersion);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);

	return true;
}

/*-----------------------------------------------

Name:   ResizeOpenGLViewportFull

Params: none

Result: Resizes viewport to full window with perspective
        projection.

/*---------------------------------------------*/

void COpenGLControl::ResizeOpenGLViewportFull()
{
	int w, h;
	if (window) {
		SDL_GetWindowSize(window, &w, &h);
		glViewport(0, 0, w, h);
	}
}

/*-----------------------------------------------

Name:   SwapBuffersM

Params: none

Result: Swaps back and front buffer.

/*---------------------------------------------*/

void COpenGLControl::SwapBuffersM()
{
	SDL_GL_SwapWindow(window);
}

/*-----------------------------------------------

Name:   Render

Params: lpParam - pointer to whatever you want

Result: Calls previously set render function.

/*---------------------------------------------*/

void COpenGLControl::Render(void * lpParam)
{
	if(RenderScene)
		RenderScene(lpParam);
}

/*-----------------------------------------------

Name:   ReleaseOpenGLControl

Params: lpParam - pointer to whatever you want

Result: Calls previously set release function
        and deletes rendering context.

/*---------------------------------------------*/

void COpenGLControl::ReleaseOpenGLControl(void * lpParam)
{
	if(ReleaseScene)
		ReleaseScene(lpParam);

	SDL_GL_DeleteContext(context);
	SDL_DestroyWindow(window);
	SDL_Quit();
	window = NULL;
}
