#include "common_header.h"
#include "Shaders.h"

CShader::CShader()
{
	bLoaded = false;
}

/*-----------------------------------------------

Name:   loadShader

Params: sFile - path to a file
        a_iType - type of shader (fragment, vertex, geometry)

Result: Loads and compiles shader.

/*---------------------------------------------*/

bool CShader::LoadShader(std::string sFile, int a_iType)
{
	FILE* fp = fopen(sFile.c_str(), "rt");
	if(!fp) {
		PRINT_ERROR("Could not load file " << sFile);
		return false;
	}

	// Get all lines from a file

	std::vector<std::string> sLines;
	char sLine[255];
	while(fgets(sLine, 255, fp))sLines.push_back(sLine);
	fclose(fp);

	const char** sProgram = new const char*[ESZ(sLines)];
	FOR(i, ESZ(sLines))sProgram[i] = sLines[i].c_str();

	uiShader = glCreateShader(a_iType);

	glShaderSource(uiShader, ESZ(sLines), sProgram, NULL);
	glCompileShader(uiShader);

	delete[] sProgram;

	int iCompilationStatus;
	glGetShaderiv(uiShader, GL_COMPILE_STATUS, &iCompilationStatus);

	if(iCompilationStatus == GL_FALSE) {
		PRINT_ERROR("Could not compile shader from file " << sFile);
		char buffer[4096];
		glGetShaderInfoLog(uiShader, 4096, NULL, buffer);
		PRINT_ERROR("OpenGL Shader Info Log:\n" << buffer);
		return false;
	}

	iType = a_iType;
	bLoaded = true;

	return 1;
}

/*-----------------------------------------------

Name:   IsLoaded

Params: none

Result: True if shader was loaded and compiled.

/*---------------------------------------------*/

bool CShader::IsLoaded()
{
	return bLoaded;
}

/*-----------------------------------------------

Name:   GetShaderID

Params: none

Result: Returns ID of a generated shader.

/*---------------------------------------------*/

unsigned int CShader::GetShaderID()
{
	return uiShader;
}

/*-----------------------------------------------

Name:   DeleteShader

Params: none

Result: Deletes shader and frees memory in GPU.

/*---------------------------------------------*/

void CShader::DeleteShader()
{
	if(!IsLoaded())return;
	bLoaded = false;
	glDeleteShader(uiShader);
}

CShaderProgram::CShaderProgram()
{
	bLinked = false;
}

/*-----------------------------------------------

Name:   CreateProgram

Params: none

Result: Creates a new program.

/*---------------------------------------------*/

void CShaderProgram::CreateProgram()
{
	uiProgram = glCreateProgram();
}

/*-----------------------------------------------

Name:   AddShaderToProgram

Params: sShader - shader to add

Result: Adds a shader (like source file) to
        a program, but only compiled one.

/*---------------------------------------------*/

bool CShaderProgram::AddShaderToProgram(CShader* shShader)
{
	if(!shShader->IsLoaded())return false;

	glAttachShader(uiProgram, shShader->GetShaderID());

	return true;
}

/*-----------------------------------------------

Name:   LinkProgram

Params: none

Result: Performs final linkage of OpenGL program.

/*---------------------------------------------*/

bool CShaderProgram::LinkProgram()
{
	glLinkProgram(uiProgram);
	int iLinkStatus;
	glGetProgramiv(uiProgram, GL_LINK_STATUS, &iLinkStatus);
	bLinked = iLinkStatus == GL_TRUE;
	return bLinked;
}

/*-----------------------------------------------

Name:   DeleteProgram

Params: none

Result: Deletes program and frees memory on GPU.

/*---------------------------------------------*/

void CShaderProgram::DeleteProgram()
{
	if(!bLinked)return;
	bLinked = false;
	glDeleteProgram(uiProgram);
}

/*-----------------------------------------------

Name:   UseProgram

Params: none

Result: Tells OpenGL to use this program.

/*---------------------------------------------*/

void CShaderProgram::UseProgram()
{
	if(bLinked)glUseProgram(uiProgram);
}

/*-----------------------------------------------

Name:   GetProgramID

Params: none

Result: Returns OpenGL generated shader program ID.

/*---------------------------------------------*/

unsigned int CShaderProgram::GetProgramID()
{
	return uiProgram;
}
