#include "common_header.h"
#include "OpenGLApp.h"
#include "Shaders.h"
#include "Texture.h"
#include "VertexBufferObject.h"
#include "WalkingCamera.h"
#include "FreeTypeFont.h"
#include "static_geometry.h"

#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

/* One VBO, where all static data are stored now,
in this tutorial vertex is stored as 3 floats for
position, 2 floats for texture coordinate and
3 floats for normal vector. */

CVertexBufferObject vboSceneObjects;
UINT uiVAOs[1]; // Only one VAO now

CShader shShaders[5];
CShaderProgram spDirectionalLight, spFont2D;

#define NUMTEXTURES 1

CTexture tTextures[NUMTEXTURES];

CWalkingCamera cCamera;

CFreeTypeFont ftFont;

/*-----------------------------------------------

Name:   InitScene

Params: lpParam - Pointer to anything you want.

Result: Initializes OpenGL features that will
        be used.

/*---------------------------------------------*/

int iTorusFaces;

void InitScene(LPVOID lpParam)
{
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

	vboSceneObjects.CreateVBO();
	glGenVertexArrays(1, uiVAOs); // Create one VAO
	glBindVertexArray(uiVAOs[0]);

	vboSceneObjects.BindVBO();
	iTorusFaces = GenerateTorus(vboSceneObjects, 7.0f, 2.0f, 20, 20);
	vboSceneObjects.UploadDataToGPU(GL_STATIC_DRAW);

	// Vertex positions
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 2*sizeof(glm::vec3)+sizeof(glm::vec2), 0);
	// Texture coordinates
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 2*sizeof(glm::vec3)+sizeof(glm::vec2), (void*)sizeof(glm::vec3));
	// Normal vectors
	glEnableVertexAttribArray(2);
	glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 2*sizeof(glm::vec3)+sizeof(glm::vec2), (void*)(sizeof(glm::vec3)+sizeof(glm::vec2)));

	// Load shaders and create shader programs

	shShaders[0].LoadShader("data/shaders/shader.vert", GL_VERTEX_SHADER);
	shShaders[1].LoadShader("data/shaders/shader.frag", GL_FRAGMENT_SHADER);
	shShaders[2].LoadShader("data/shaders/ortho2D.vert", GL_VERTEX_SHADER);
	shShaders[4].LoadShader("data/shaders/font2D.frag", GL_FRAGMENT_SHADER);

	spDirectionalLight.CreateProgram();
	spDirectionalLight.AddShaderToProgram(&shShaders[0]);
	spDirectionalLight.AddShaderToProgram(&shShaders[1]);
	spDirectionalLight.LinkProgram();

	spFont2D.CreateProgram();
	spFont2D.AddShaderToProgram(&shShaders[2]);
	spFont2D.AddShaderToProgram(&shShaders[4]);
	spFont2D.LinkProgram();

	// Load textures

	std::string sTextureNames[] = {"rust.jpg"};

	FOR(i, NUMTEXTURES) // I know that FOR cycle is useless now, but it was easier to rewrite :)
	{
		tTextures[i].LoadTexture2D("data/textures/" + sTextureNames[i], true);
		tTextures[i].SetFiltering(TEXTURE_FILTER_MAG_BILINEAR, TEXTURE_FILTER_MIN_BILINEAR_MIPMAP);
	}

	glEnable(GL_DEPTH_TEST);
	glClearDepth(1.0);

	// Here we load font with pixel size 32 - this means that if we print with size above 32, the quality will be low
	ftFont.LoadFont("data/fonts/ComicNeue_Bold.otf", 32);
//	ftFont.LoadFont("data/fonts/ComicNeue-Regular.otf", 32);
	ftFont.SetShaderProgram(&spFont2D);

	cCamera = CWalkingCamera(glm::vec3(0.0f, 1.0f, -30.0f), glm::vec3(0.0f, 1.0f, -29.0f), glm::vec3(0.0f, 1.0f, 0.0f), 35.0f);

	if (!CheckGLError("InitScene"))
		PRINT_INFO("Initialized scene with no errors");
}

/*-----------------------------------------------

Name:   RenderScene

Params: lpParam - Pointer to anything you want.

Result: Renders whole scene.

/*---------------------------------------------*/

float fGlobalAngle; // In radians now

int iFontSize = 24; // This is default printed font size

void RenderScene(LPVOID lpParam)
{
	// Typecast lpParam to COpenGLControl pointer
	COpenGLControl* oglControl = (COpenGLControl*)lpParam;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	spDirectionalLight.UseProgram();
	glBindVertexArray(uiVAOs[0]);
	// Set some lighting parameters

	spDirectionalLight.SetUniform("sunLight.vColor", glm::vec3(1.0f, 1.0f, 1.0f));
	spDirectionalLight.SetUniform("sunLight.fAmbientIntensity", 0.25f);
	spDirectionalLight.SetUniform("sunLight.vDirection", glm::vec3(0, -1, 0));

	spDirectionalLight.SetUniform("projectionMatrix", oglControl->GetProjectionMatrix());

	glm::mat4 mModelView = cCamera.look();
	glm::mat4 mModelToCamera = glm::rotate(glm::mat4(1.0f), fGlobalAngle, glm::vec3(0.0f, 1.0f, 0.0f));

	// Render torus
	spDirectionalLight.SetUniform("gSampler", 0);
	spDirectionalLight.SetUniform("normalMatrix", glm::transpose(glm::inverse(mModelToCamera)));
	spDirectionalLight.SetUniform("modelViewMatrix", mModelView*mModelToCamera);
	spDirectionalLight.SetUniform("vColor", glm::vec4(1.0f, 1.0f, 1.0f, 1.0f));

	tTextures[0].BindTexture();
	glDrawArrays(GL_TRIANGLES, 0, iTorusFaces*3);

	glDisable(GL_DEPTH_TEST);
	glDepthFunc(GL_ALWAYS);

	spFont2D.UseProgram();
	// Font color, you can even change transparency of font with alpha parameter
	spFont2D.SetUniform("vColor", glm::vec4(1.0f, 1.0f, 1.0f, 1.0f));
	spFont2D.SetUniform("projectionMatrix", oglControl->GetOrthoMatrix());

	char buf[255];
	sprintf(buf, "Font Size: %d\nPress UP and DOWN arrow key to change\n\nTotal Torus Faces: %d", iFontSize, iTorusFaces);

	ftFont.print(buf, 20, oglControl->GetViewportHeight()-10-iFontSize, iFontSize);

	ftFont.print("www.mbsoftworks.sk", 20, 20, 24);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	fGlobalAngle += appMain.sof(glm::radians(100.0f));
	cCamera.update();

	const unsigned char * state = SDL_GetKeyboardState(NULL);
	if (state[SDL_SCANCODE_UP])
		iFontSize++;
	if (state[SDL_SCANCODE_DOWN] && iFontSize > 0)
		iFontSize--;

	oglControl->SwapBuffersM();

	CheckGLError("RenderScene");
}

/*-----------------------------------------------

Name:   ReleaseScene

Params: lpParam - Pointer to anything you want.

Result: Releases OpenGL scene.

/*---------------------------------------------*/

void ReleaseScene(LPVOID lpParam)
{
	FOR(i, NUMTEXTURES)tTextures[i].ReleaseTexture();

	spDirectionalLight.DeleteProgram();
	spFont2D.DeleteProgram();
	FOR(i, 5)shShaders[i].DeleteShader();
	ftFont.ReleaseFont();

	glDeleteVertexArrays(1, uiVAOs);
	vboSceneObjects.ReleaseVBO();
}
